#!/usr/bin/ruby

require 'cgi'
require 'logger'
require 'rubygems'
require 'json'
require 'systemu'

$DTSADMIN_CMD = '/usr/bin/dtsadmin'

$CACHE_PATH = '/tmp/server_config'

$LOG_PATH = '/tmp/web_api.log'
$LOG_LEVEL = Logger::INFO

$ERR_CODE = 999

def check_value(value)
  if value =~ /[ #;&,+*~\':"!^$%\[\]()=>|\/]/ then
    return false
  end
  return true
end

def exec_cmd(cmd)
  res = systemu(cmd)
  return res[0], res[1], res[2]
end

def exec_cmd_stdin(cmd, data)
  res = systemu(cmd, :stdin => data)
  return res[0], res[1], res[2]
end

def res_file(cgi, file)
  if file == nil then
    file = ''
  end

  filename = Time.now.strftime("%Y%m%d_%H%M%S") + '.txt'
  print cgi.header({'status' => 'OK',
                    'type' => 'application/octet-stream',
                    'Content-Disposition' => "attachment; filename=\"#{filename}\"",
                    'length' => file.size})
  print file
end

def res_load_file(cgi, data)
  body = JSON.pretty_generate(data)
  print cgi.header({'status' => 'OK',
                    'type' => 'text/html',
                    'length' => body.size})
  print body
end

def res_json(cgi, data)
  body = JSON.pretty_generate(data)
  print cgi.header({'status' => 'OK',
                    'type' => 'application/json',
                    'length' => body.size})
  print body
end

def res_err(cgi, data)
  msg = data[:msg] + '(' + data[:status].to_s + ')'
  print cgi.header({'status' => 'SERVER_ERROR',
                    'type' => 'text/html',
                    'charset' => 'iso-8859-1',
                    'length' => msg.size})
  print msg
end

def clear(cgi)
  cmd = sprintf('%s -C', $DTSADMIN_CMD)

  $logger.debug("clear cmd: #{cmd}")

  status, stdout, stderr = exec_cmd(cmd)
  if status.exited? then
    if status.exitstatus == 0 then
      return {:result => true}
    end
  end
  return {:result => false, :status => status.exitstatus, :msg => stderr.chomp}
end

def save_setting(cgi)
  cmd = sprintf('%s -S', $DTSADMIN_CMD)

  $logger.debug("save cmd: #{cmd}")

  status, stdout, stderr = exec_cmd(cmd)
  if status.exited? then
    if status.exitstatus == 0 then
      return {:result => true, :data => stdout}
    end
  end
  return {:result => false, :status => status.exitstatus, :msg => stderr.chomp}
end

def load_setting(cgi, file_data)
  if file_data == nil || file_data.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'file is null'}
  end

  cmd = sprintf('%s -R', $DTSADMIN_CMD)

  $logger.debug("load cmd: #{cmd}")

  file_data = file_data.read.gsub(/\r\n/, "\n")

  status, stdout, stderr = exec_cmd_stdin(cmd, file_data)
  if status.exited? then
    if status.exitstatus == 0 then
      return {:result => true}
    end
  end
  return {:result => false, :status => status.exitstatus, :msg => stderr.chomp}
end

def server_list(cgi)
  cmd = sprintf('%s -L -v -n -c', $DTSADMIN_CMD)

  $logger.debug("server list cmd: #{cmd}")

  status, stdout, stderr = exec_cmd(cmd)
  if status.exited? then
    if status.exitstatus == 0 then
      # write dtsadmin cache
      File.open($CACHE_PATH, 'w') do |file|
        file.puts stdout
      end

      servers = parse_server_list(stdout)
      return {:result => true, :servers => servers}
    end
  end
  return {:result => false, :status => status.exitstatus, :msg => stderr.chomp}
end

def node_status(cgi, ipaddress, port)
  if ipaddress == nil || ipaddress.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'IP Address is null'}
  elsif port == nil || port.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'Port Number is null'}
  end

  if !check_value(ipaddress) then
    msg = sprintf('IP Address is invalid value(%s)', ipaddress)
    return {:result => false, :status => $ERR_CODE, :msg => msg}
  elsif !check_value(port) then
    msg = sprintf('Port Number is invalid value(%s)', port)
    return {:result => false, :status => $ERR_CODE, :msg => msg}
  end

  cmd = sprintf('%s -L -v -n -c', $DTSADMIN_CMD)

  $logger.debug("node status cmd: #{cmd}")

  status, stdout, stderr = exec_cmd(cmd)
  if status.exited? then
    if status.exitstatus == 0 then
      # write dtsadmin cache
      File.open($CACHE_PATH, 'w') do |file|
        file.puts stdout
      end

      target_server = ipaddress + ':' + port
      servers = parse_server_list(stdout)
      servers.each do |server|
        if target_server =~ /^#{server[:ipaddress]}:#{server[:port]}$/ then
          return {:result => true, :server => server}
        end
      end

      return {:result => false, :status => $ERR_CODE, :msg => "#{target_server} data is not found"}
    end
  end
  return {:result => false, :status => status.exitstatus, :msg => stderr.chomp}
end

def update_node(cgi, type, ipaddress, port, node, vnodeids)
  if ipaddress == nil || ipaddress.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'IP Address is null'}
  elsif port == nil || port.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'Port Number is null'}
  elsif node == nil || node.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'Node Name is null'}
  end

  if !check_value(ipaddress) then
    msg = sprintf('ipaddress is invalid value(%s)', ipaddress)
    return {:result => false, :status => $ERR_CODE, :msg => msg}
  elsif !check_value(port) then
    msg = sprintf('port is invalid value(%s)', port)
    return {:result => false, :status => $ERR_CODE, :msg => msg}
  elsif !check_value(node) then
    msg = sprintf('node is invalid value(%s)', node)
    return {:result => false, :status => $ERR_CODE, :msg => 'Node Name is invalid value'}
  end

  vnodeid = ''
  if type.to_s == 'add_vnode' || type.to_s == 'del_vnode' then
    if vnodeids != nil && vnodeids.length > 0 then
      vnodeids.each do |id|
        vnodeid += id
        vnodeid += ','
      end
    end
  end

  vnodeid = vnodeid.sub(/,$/, '')

  cmd = nil
  if type.to_s == 'add_vnode' then
    if vnodeid == nil || vnodeid.length < 1 then
      cmd = sprintf('%s -a %s:%s %s', $DTSADMIN_CMD, ipaddress, port, node)
    else
      cmd = sprintf('%s -a %s:%s %s/%s', $DTSADMIN_CMD, ipaddress, port, node, vnodeid)
    end
  elsif type.to_s == 'del_vnode' then
    if vnodeid == nil || vnodeid.length < 1 then
      cmd = sprintf('%s -d %s:%s %s', $DTSADMIN_CMD, ipaddress, port, node)
    else
      cmd = sprintf('%s -d %s:%s %s/%s', $DTSADMIN_CMD, ipaddress, port, node, vnodeid)
    end
  elsif type.to_s == 'del_node' then
    cmd = sprintf('%s -d %s:%s %s', $DTSADMIN_CMD, ipaddress, port, node)
  else
    return {:result => false, :status => $ERR_CODE, :msg => 'invalid type'}
  end

  $logger.debug("update node cmd: #{cmd}")

  status, stdout, stderr = exec_cmd(cmd)
  if status.exited? then
    if status.exitstatus == 0 then
      return {:result => true}
    end
  end
  return {:result => false, :status => status.exitstatus, :msg => stderr.chomp}
end

def add_node(cgi, ipaddress, port, node, add_vnodeid)
  return update_node(cgi, :add_vnode, ipaddress, port, node, add_vnodeid)
end

def apply_node(cgi, ipaddress, port, node, add_vnodeid, del_vnodeid)
  if add_vnodeid != nil && add_vnodeid.length > 0 then
    result = update_node(cgi, :add_vnode, ipaddress, port, node, add_vnodeid)
    if !result[:result] then
      return result
    end
  end

  if del_vnodeid != nil && del_vnodeid.length > 0 then
    result = update_node(cgi, :del_vnode, ipaddress, port, node, del_vnodeid)
    if !result[:result] then
      return result
    end
  end

  return {:result => true}
end

def delete_node(cgi, ipaddress, port, node)
  return update_node(cgi, :del_node, ipaddress, port, node, nil)
end

def add_server_setting(cgi, ipaddress, port, if_name, netmask, fakemac, ri_name, ri_option, nodes)
  if ipaddress == nil || ipaddress.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'IP Address is null'}
  elsif port == nil || port.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'Port Number is null'}
  elsif if_name == nil || if_name.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'Iinterface Name is null'}
  elsif ri_name == nil || ri_name.length < 1 then
    return {:result => false, :status => $ERR_CODE, :msg => 'Request Inspector Name is null'}
  end

  if !check_value(ipaddress) then
    msg = sprintf('IP Address is invalid value(%s)', ipaddress)
    return {:result => false, :status => $ERR_CODE, :msg => msg}
  elsif !check_value(port) then
    msg = sprintf('Port Number is invalid value(%s)', port)
    return {:result => false, :status => $ERR_CODE, :msg => msg}
  elsif !check_value(if_name) then
    msg = sprintf('Interface Name is invalid value(%s)', if_name)
    return {:result => false, :status => $ERR_CODE, :msg => 'Interface Name is invalid value'}
  elsif !check_value(ri_name) then
    msg = sprintf('Request Inspector Name is invalid value(%s)', ri_name)
    return {:result => false, :status => $ERR_CODE, :msg => 'Request Inspector Name is invalid value'}
  end

  fakemac_str = ''
  if fakemac then
    fakemac_str = 'on'
  else
    fakemac_str = 'off'
  end

  nodes_str = ''
  nodes.each do |node, ids|
    nodes_str += node

    add_delim = false
    for i in 1..63
      if ids[i] == 'true' then
        if !add_delim then
          nodes_str += '/'
          add_delim = true
        end
        nodes_str += i.to_s
        nodes_str += ','
      end
    end
    nodes_str += ':'
  end

  nodes_str = nodes_str.gsub(/,:/, ':').gsub(/,$|:$/, '')

  cmd = nil
  if netmask == nil || netmask.length < 1 then
    cmd = sprintf('%s -A %s:%s -i %s -f %s -r %s -s "%s" %s', $DTSADMIN_CMD,
                                                              ipaddress,
                                                              port,
                                                              if_name,
                                                              fakemac_str,
                                                              ri_name,
                                                              ri_option,
                                                              nodes_str)
  else
    cmd = sprintf('%s -A %s:%s -i %s -M %s -f %s -r %s -s "%s" %s', $DTSADMIN_CMD,
                                                                    ipaddress,
                                                                    port,
                                                                    if_name,
                                                                    netmask,
                                                                    fakemac_str,
                                                                    ri_name,
                                                                    ri_option,
                                                                    nodes_str)
  end

  $logger.debug("add server setting cmd: #{cmd}")

  status, stdout, stderr = exec_cmd(cmd)
  if status.exited? then
    if status.exitstatus == 0 then
      return {:result => true}
    end
  end
  return {:result => false, :status => status.exitstatus, :msg => stderr.chomp}
end

def parse_server_list(str)
  start = false
  servers = Array.new
  host = nil
  str.each do |line|
    if start then
      if line =~ /(\S+)\:(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+"(.*)"\s+(\S+)$/ then
        host = {:ipaddress => $1,
                :port => $2,
                :interface => $3,
                :netmask => $4,
                :fakemac => $5,
                :ri => $7,
                :ri_opt => $8,
                :proxyport => $9,
                :nodes => Array.new}
        servers.push(host)
      elsif line =~ /^\s+\-\>\s+(\S+)\(dead\)$/ then
        nodes = host[:nodes]
        nodes.push({:hbname => $1, :ready => false})
      elsif line =~ /^\s+\-\>\s+(\S+)$/ then
        nodes = host[:nodes]
        nodes.push({:hbname => $1, :ready => true})
      elsif line =~ /^\s+\-\>\s+(\S+)\s+(\S+)\s+(\S+)$/ then
        nodes = host[:nodes]
        nodes.push({:hbname => $1, :vnodeid => $2, :hashvalue => $3, :ready => true})
      else
        $logger.warn("parse warn: #{$line}")
      end
    elsif line =~ /^\s+\-\> HeartbeatHostName \[VirtualNodeID HashValue\]/ then
      start = true
    end
  end

  return servers
end

def result_to_str(result)
  str = ''
  result.each do |key, value|
    if value.class.name == 'Hash' || value.class.name == 'Array' then
      sub_str = result_to_str(value)
      str += sprintf('%s: [%s], ', key.to_s, sub_str.to_s)
    else
      str += sprintf('%s: %s, ', key.to_s, value.to_s)
    end
  end
  return str
end

#
# main
#
$logger = Logger.new($LOG_PATH)
$logger.level = $LOG_LEVEL

cgi = CGI.new

service = nil
content_type = cgi.content_type
if content_type =~ /^multipart\/form-data/ then
  service = cgi.params['service'][0].read
else
  service = cgi.params['service'][0]
end

case service
when 'clear' then
  result = clear(cgi)

  if $logger.debug? then
    $logger.debug("service: clear, " + result_to_str(result))
  end

  res_json(cgi, result)
when 'save_setting' then
  result = save_setting(cgi)

  if $logger.debug? then
    $logger.debug("service: save_setting, " + result_to_str(result))
  end

  if result[:result] then
    res_file(cgi, result[:data])
  else
    res_err(cgi, result)
  end
when 'load_setting' then
  result = load_setting(cgi, cgi.params['load_file'][0])

  if $logger.debug? then
    $logger.debug("service: load_setting, " + result_to_str(result))
  end

  res_load_file(cgi, result)
when 'server_list' then
  result = server_list(cgi)

  if $logger.debug? then
    $logger.debug("service: server_list, " + result_to_str(result))
  end

  res_json(cgi, result)
when 'node_status' then
  ipaddress = cgi.params['ipaddress'][0]
  port = cgi.params['port'][0]
  result = node_status(cgi, ipaddress, port)

  if $logger.debug? then
    $logger.debug("service: node_status, " + result_to_str(result))
  end

  res_json(cgi, result)
when 'add_node' then
  ipaddress = cgi.params['ipaddress'][0]
  port = cgi.params['port'][0]
  node = cgi.params['node'][0]
  add_vnodeid = cgi.params['add_vnodeid[]'].to_a
  result = add_node(cgi, ipaddress, port, node, add_vnodeid)

  if $logger.debug? then
    $logger.debug("service: add_node, " + result_to_str(result))
  end

  res_json(cgi, result)
when 'delete_node' then
  ipaddress = cgi.params['ipaddress'][0]
  port = cgi.params['port'][0]
  node = cgi.params['node'][0]
  result = delete_node(cgi, ipaddress, port, node)

  if $logger.debug? then
    $logger.debug("service: delete_node, " + result_to_str(result))
  end

  res_json(cgi, result)
when 'apply_node' then
  ipaddress = cgi.params['ipaddress'][0]
  port = cgi.params['port'][0]
  node = cgi.params['node'][0]
  add_vnodeid = cgi.params['add_vnodeid[]'].to_a
  del_vnodeid = cgi.params['del_vnodeid[]'].to_a
  result = apply_node(cgi, ipaddress, port, node, add_vnodeid, del_vnodeid)

  if $logger.debug? then
    $logger.debug("service: apply_node, " + result_to_str(result))
  end

  res_json(cgi, result)
when 'add_server_setting' then
  ipaddress = cgi.params['ipaddress'][0]
  port = cgi.params['port'][0]
  if_name = cgi.params['if_name'][0]
  netmask = cgi.params['netmask'][0]
  fakemac = cgi.params['fakemac'][0]
  ri_name = cgi.params['ri_name'][0]
  ri_option = cgi.params['ri_option'][0]
  host_name = cgi.params['host_name[]'].to_a
  nodes = Hash.new
  host_name.each do |host|
    id = cgi.params[host + '_vnodeid[]'].to_a
    nodes[host] = id
  end
  result = add_server_setting(cgi, ipaddress, port, if_name, netmask,
                              fakemac, ri_name, ri_option, nodes)

  if $logger.debug? then
    $logger.debug("service: add_server_setting, " + result_to_str(result))
  end

  res_json(cgi, result)
else
  result = {:result => false, :status => $ERR_CODE, :msg => 'invalid service'}

  if $logger.debug? then
    $logger.debug("service: invalid service, " + result_to_str(result))
  end

  res_json(cgi, result)
end

